/*
 * Copyright (c) 2004-2005 Northwestern University.
 *                         NuCAD Group.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/*
 * acg_base.cpp - Implementation details of acg_base.h
 *
 * Authors: Hai Zhou, haizhou@ece.northwestern.edu
 *          Jia Wang, jwa112@ece.northwestern.edu
 */

#include "../misc/common.h"

#include "acg_base.h"
#include "acg_dev.h"

namespace acg
{

bool edge_t::valid() const
{
	for (int i = 0; i < 2; ++i)
	{
		pos_t pos = pos_t(i);

		if (next(pos)->prev(pos) != this)
			return false;
		if (prev(pos)->next(pos) != this)
			return false;

		if (node(pos) != next(pos)->node(pos))
			return false;
		if (node(pos) != prev(pos)->node(pos))
			return false;
	}

	return true;
}

edge_t *edge_list_t::pop_front(pos_t pos)
{
	edge_t *tmp = front(pos);
	erase(tmp, pos);
	return tmp;
}

edge_t *edge_list_t::pop_back(pos_t pos)
{
	edge_t *tmp = back(pos);
	erase(tmp, pos);
	return tmp;
}

void edge_list_t::push_front(edge_t *edge, pos_t pos)
{
	insert(head_.next(pos), edge, pos);
}

void edge_list_t::push_back(edge_t *edge, pos_t pos)
{
	insert(&head_, edge, pos);
}

void edge_list_t::insert(edge_t *after, edge_t *edge, pos_t pos)
{
	assert(after->valid());
	assert(edge->valid());

	assert(edge->node_[pos] == 0);

	edge->node_[pos] = after->node_[pos];

	edge->prev_[pos] = after->prev_[pos];
	edge->next_[pos] = after;
	after->prev_[pos]->next_[pos] = edge;
	after->prev_[pos] = edge;
}

void edge_list_t::erase(edge_t *edge, pos_t pos)
{
	assert(edge->valid());

	assert(edge->node_[pos] != 0);
	assert(edge != edge->next_[pos]);

	edge->node_[pos] = 0;

	edge->prev_[pos]->next_[pos] = edge->next_[pos];
	edge->next_[pos]->prev_[pos] = edge->prev_[pos];
	edge->prev_[pos] = edge->next_[pos] = edge;
}

void edge_list_t::splice(edge_t *after, edge_t *first, edge_t *last, pos_t pos)
{
	assert(after->valid());
	assert(first->valid());
	assert(last->valid());

	assert(after->node_[pos] != 0);
	assert(first->node(pos) == last->node(pos));

	if (first == last)
		return;
	
	// worth splice, relink
	edge_t *prev_first = first->prev_[pos];
	edge_t *prev_after = after->prev_[pos];
	edge_t *prev_last = last->prev_[pos];

	// prev_after<->after, prev_first<->first...prev_last<->last
	// prev_after<->first...prev_last<->after, prev_first<->last

	prev_after->next_[pos] = first;
	first->prev_[pos] = prev_after;

	prev_last->next_[pos] = after;
	after->prev_[pos] = prev_last;

	prev_first->next_[pos] = last;
	last->prev_[pos] = prev_first;

	// reset node
	for (; first != after; first = first->next(pos))
		first->node_[pos] = after->node_[pos];
}

bool edge_list_t::valid() const
{
	if (!head_.valid())
		return false;

	for (int i = 0; i < 2; ++i)
	{
		pos_t pos = pos_t(i);

		for (const_range_t r = all(pos);
			r.first != r.second;
			r.first = r.first->next(pos))
			if (!r.first->valid())
				return false;
	}

	if (head_.node(POS_IN) != head_.node(POS_OUT))
		return false;

	return true;
}

node_t::node_t() : m_id(0), m_data(0)
{
	edges_[0].init(this);
	edges_[1].init(this);

	width_[0] = width_[1] = 0;
	bb_[0][0] = bb_[0][1] = bb_[1][0] = bb_[1][1] = 0;
}

node_t::node_t(double w, double h, size_t id)
	: m_id(id), m_data(0)
{
	edges_[0].init(this);
	edges_[1].init(this);

	width_[0] = w; width_[1] = h;
	bb_[0][0] = bb_[0][1] = bb_[1][0] = bb_[1][1] = 0;
}

bool node_t::valid() const
{
	if (!edges(COLOR_RED)->valid())
		return false;
	if (!edges(COLOR_BLACK)->valid())
		return false;

	if (edges(COLOR_RED)->node() != this)
		return false;
	if (edges(COLOR_BLACK)->node() != this)
		return false;

	return true;
}

void node_t::copy_content(const node_t &src)
{
	m_id = src.m_id;
	memcpy(width_, src.width_, sizeof(width_));
	memcpy(bb_, src.bb_, sizeof(bb_));
}

void node_t::swap_content(node_t &src)
{
	std::swap(m_id, src.m_id);
	std::swap(width_[0], src.width_[0]);
	std::swap(width_[1], src.width_[1]);
	std::swap(bb_[0][0], src.bb_[0][0]);
	std::swap(bb_[0][1], src.bb_[0][1]);
	std::swap(bb_[1][0], src.bb_[1][0]);
	std::swap(bb_[1][1], src.bb_[1][1]);
}

acg_base_t::acg_base_t()
	: n_edges_(0)
{
}

acg_base_t::acg_base_t(size_t n, misc::dbl_pair shapes[], size_t ids[])
	: n_edges_(0), nodes_(n)
{
	acg_dev_t dev(*this);

	for (size_t i = 0; i < n; ++i)
		nodes_[i] = dev.alloc_node(
			shapes[i].first, shapes[i].second,
			(ids == 0)? i: ids[i]);
}

acg_base_t::acg_base_t(const acg_base_t &r)
	: n_edges_(0), nodes_(r.num_nodes())
{
	acg_dev_t dev(*this);

	for (size_t i = 0; i < r.num_nodes(); ++i)
	{
		nodes_[i] = dev.alloc_node();
		nodes_[i]->copy_content(*r[i]);
	}

	dev.duplicate_edges(r);
}

acg_base_t::~acg_base_t()
{
	clear();
}

void acg_base_t::clear()
{
	acg_dev_t dev(*this);

	dev.free_all_edges();

	for (size_t i = 0; i < num_nodes(); ++i)
		dev.free_node(nodes_[i]);
	nodes_.clear();
}

void acg_base_t::swap(acg_base_t &r)
{
	nodes_.swap(r.nodes_);
	std::swap(n_edges_, r.n_edges_);
}

void acg_base_t::node_swap(size_t i, size_t j)
{
	at(i)->swap_content(*at(j));
}

void acg_base_t::node_rotate(size_t where)
{
	node_t *node = at(where);
	std::swap(node->width(COLOR_RED), node->width(COLOR_BLACK));
}

void acg_base_t::node_resize(size_t where, double w, double h)
{
	// could be useful for soft modules
	node_t *node = at(where);
	node->width(COLOR_RED) = w;
	node->width(COLOR_BLACK) = h;
}


namespace detail
{

void xml_write_node_edges(FILE *fp, const node_t *node, color_t c)
{
	const char *color = (c == COLOR_RED)? "red": "black";

	for (edge_list_t::const_range_t r = node->edges(c)->all(POS_IN);
		r.first != r.second; r.first = r.first->next(POS_IN))
		fprintf(fp,
			"    <edge color='%s' from='%p' to='%p' />\n",
			color, node, r.first->node(POS_OUT));
}

} // namespace detail

void acg_base_t::xml_write(FILE *fp, const char *type)
{
	using detail::xml_write_node_edges;

	const size_t n = num_nodes();

	fprintf(fp, "  <modules size='%u'>\n", n);
	for (size_t i = 0; i < n; ++i)
	{
		fprintf(fp,
			"    <module name='%p' id='%u' seq='%u'\n"
			"      x='%e' y='%e' width='%e' height='%e'\n"
			"      llx='%e' lly='%e' urx='%e' ury='%e' />\n",
			at(i), at(i)->m_id, i,
			at(i)->bb(acg::COLOR_RED, acg::POS_IN),
			at(i)->bb(acg::COLOR_BLACK, acg::POS_IN),
			at(i)->width(acg::COLOR_RED),
			at(i)->width(acg::COLOR_BLACK),
			at(i)->bb(acg::COLOR_RED, acg::POS_IN),
			at(i)->bb(acg::COLOR_BLACK, acg::POS_IN),
			at(i)->bb(acg::COLOR_RED, acg::POS_OUT),
			at(i)->bb(acg::COLOR_BLACK, acg::POS_OUT));
	}
	fprintf(fp, "  </modules>\n");

	fprintf(fp, "  <edges type='%s' size='%u'>\n", type, num_edges());
	for (size_t i = 0; i < n; ++i)
	{
		xml_write_node_edges(fp, at(i), COLOR_RED);
		xml_write_node_edges(fp, at(i), COLOR_BLACK);
	}
	fprintf(fp, "  </edges>\n");
}

} // namespace acg
