/*
 * Copyright (c) 2004-2005 Northwestern University.
 *                         NuCAD Group.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/*
 * rcg.h - Implement algorithms for RCG.
 *
 * Authors: Hai Zhou, haizhou@ece.northwestern.edu
 *          Jia Wang, jwa112@ece.northwestern.edu
 */

#include "../misc/common.h"

#include "rcg.h"
#include "acg_racg.h"
#include "acg_dev.h"

namespace acg
{

namespace detail
{

color_t node_color(const node_t *node, pos_t pos); // from acg_racg.cpp

void rcg_remove_transitive(acg_dev_t &base,
	node_t *nA, node_t *nB, pos_t pos, color_t c);

void rcg_add_original(acg_dev_t &base,
	node_t *nA, node_t *nB, pos_t pos, color_t c, edge_t *eAB);

} // namespace detail

void rcg_build(acg_base_t &base)
{
	racg_build(base);

	racg_acg_context_t ctx;

	racg_to_acg(base, ctx);
}

void rcg_swap(acg_base_t &b, size_t where)
{
	using detail::node_color;
	using detail::rcg_remove_transitive;
	using detail::rcg_add_original;

	if (where == b.num_nodes()-1)
		return;

	acg_dev_t base(b);

	node_t *node0 = base[where];
	node_t *node1 = base[where+1];

	color_t c = node_color(node0, POS_IN);
	assert(c == node_color(node1, POS_OUT));
	assert(node0 == node1->nearest_node(c, POS_OUT));
	assert(node1 == node0->nearest_node(c, POS_IN));

	rcg_remove_transitive(base, node0, node1, POS_OUT, other(c));
	rcg_remove_transitive(base, node1, node0, POS_IN, other(c));

	edge_t *change = node0->edges(c)->front(POS_IN);
	assert(change->node(POS_OUT) == node1);

	rcg_add_original(base, node0, node1, POS_OUT, c, change);
	rcg_add_original(base, node1, node0, POS_IN, c, change);

	// re-assign change
	edge_list_t::erase(change, POS_IN);
	edge_list_t::erase(change, POS_OUT);
	node0->edges(other(c))->push_front(change, POS_OUT);
	node1->edges(other(c))->push_front(change, POS_IN);
	
	// swap the node
	std::swap(base[where], base[where+1]);
}

namespace detail
{

/*
 * A--B      A
 * :  :      :
 * ?  :  =>  B
 * :  :      :
 * CCCC      C
 */
void rcg_remove_transitive(acg_dev_t &base,
	node_t *nA, node_t *nB, pos_t pos, color_t c)
{
	assert(nB->nearest_node(other(c), pos) == nA);
	assert(nA->nearest_node(other(c), other(pos)) == nB);

	for (edge_list_t::range_t rB = nB->edges(c)->all(pos);
		rB.first != rB.second; rB.first = rB.first->next(pos))
	{
		edge_t *eCB = rB.first;
		node_t *nC = eCB->node(other(pos)); (void)nC;

		// there is a c path from C to A
		assert(eCB != nC->edges(c)->front(other(pos)));

		edge_t *eCA = eCB->prev(other(pos));
		if (eCA->node(pos) != nA) // possiblely not eCA
			break;

		// remove eCA
		edge_list_t::erase(eCA, POS_IN);
		edge_list_t::erase(eCA, POS_OUT);
		base.free_edge(eCA);
	}
}

/*
 * C--A--B      C-----A
 * C  :  B  =>  C     :
 * C--D??B      C--D  B
 */
void rcg_add_original(acg_dev_t &base,
	node_t *nA, node_t *nB, pos_t pos, color_t c, edge_t *eAB)
{
	assert(eAB->node(pos) == nB);
	assert(eAB->node(other(pos)) == nA);

	for (edge_list_t::range_t rA = nA->edges(c)->all(pos);
		rA.first != rA.second; rA.first = rA.first->next(pos))
	{
		edge_t *eCA = rA.first;
		node_t *nC = eCA->node(other(pos));

		if (eCA != nC->edges(c)->front(other(pos)))
		{
			node_t *nD = eCA->prev(other(pos))->node(pos);

			// oc path from D to B iff eDB is oc

			/* this won't work since eDA is erased if eDB is there :(
			// looking for eDA
			edge_list_t::range_t rD = nD->edges(other(c))->all(other(pos));
			for (;; rD.first = rD.first->next(other(pos)))
			{
				assert(rD.first != rD.second);

				if (rD.first->node(pos) == nA)
					break;
			}
			edge_t *eDA = rD.first;

			// if eDB is oc, it must be the one after eDA
			// should get out of the loop if eDB is not there
			// since there must be a c path from D to B
			if (eDA->next(other(pos))->node(pos) != nB)
				break;
			*/

			edge_list_t::range_t rD = nD->edges(other(c))->all(other(pos));
			for (; rD.first != rD.second; rD.first = rD.first->next(other(pos)))
				if (rD.first->node(pos) == nB)
					break;
			if (rD.first == rD.second) // eDB is not there
				break;
		}

		// add edge from C to B
		edge_t *eCB = base.alloc_edge();
		edge_list_t::insert(eCA, eCB, other(pos));
		edge_list_t::insert(eAB, eCB, pos);
	}
}

} // namespace detail

} // namespace acg
