/*
 * Copyright (c) 2004-2005 Northwestern University.
 *                         NuCAD Group.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/*
 * annealing.cpp - SA.
 *
 * Authors: Hai Zhou, haizhou@ece.northwestern.edu
 *          Jia Wang, jwa112@ece.northwestern.edu
 */

#include "../misc/common.h"

#include "annealing.h"

namespace fplan
{

namespace detail
{

class anneal_t
{
protected:
	cost_ptr &cost_;
	perturb_ptr &pt_;
	FILE * const fp_;

	const int N_, k_;

	double best_cost_, prev_cost_;
	int counter_;

	double initial_temperature();

	// return reject rate
	double round(double T, int n, double &avg);

public:
	anneal_t(cost_ptr &cost, perturb_ptr &pt, int N, FILE *fp, int k);

	int go(double r, double term_T, double max_rr);

}; // class anneal_t

class greedy_t
{
protected:
	cost_ptr &cost_;
	perturb_ptr &pt_;
	FILE * const fp_;

	const int N_;

	double best_cost_;
	int counter_;

	// return reject rate
	double round(int n);

public:
	greedy_t(cost_ptr &cost, perturb_ptr &pt, int N, FILE *fp);

	int go(double max_rr);

	int go_time(double sec);

}; // class greedy_t

} // namespace detail

int anneal(cost_ptr c, perturb_ptr pt, int N, FILE *fp, int k,
	double r, double term_T, double max_reject_rate)
{
	detail::anneal_t an(c, pt, N, fp, k);

	return an.go(r, term_T, max_reject_rate);
}

int greedy(cost_ptr c, perturb_ptr pt, int N, FILE *fp, double max_reject_rate)
{
	detail::greedy_t gr(c, pt, N, fp);

	return gr.go(max_reject_rate);
}

int greedy_time(cost_ptr c, perturb_ptr pt, int N, FILE *fp, double sec)
{
	detail::greedy_t gr(c, pt, N, fp);

	return gr.go_time(sec);
}

namespace detail
{

anneal_t::anneal_t(cost_ptr &cost, perturb_ptr &pt, int N, FILE *fp, int k)
	: cost_(cost), pt_(pt), fp_(fp), N_(N), k_(k),
	best_cost_(0), prev_cost_(0), counter_(0)
{
}

int anneal_t::go(double r, double term_T, double max_rr)
{
	clock_t s = clock();

	// accept initial solution
	prev_cost_ = best_cost_ = cost_->cost();
	pt_->record_best();
	if (fp_ != 0) fprintf(fp_, "initial cost %.3f\n", best_cost_);

	// initial temperature
	counter_ = 0;
	double T = initial_temperature();

	// annealing
	for (int iter = 1;; ++iter)
	{
		if (fp_ != 0) fprintf(fp_, "iteration %d, T = %f\n", iter, T);
	
		double avg = 0;
		double rr = round(T, N_*k_, avg);

		if (rr >= max_rr)
		{
			if (rr >= 0.95)
				if (fp_ != 0) fprintf(fp_, "convergent\n");
			break;
		}
		else if (T <= term_T)
		{
			if (fp_ != 0) fprintf(fp_, "cool enough, avg = %f, T=%f\n", avg, T);
			break;
		}

		T *= r;
	}

	pt_->swap_best();

	clock_t ss = clock();

	if (fp_ != 0) fprintf(fp_,
		"final %.3f, counter %d, time %f\n",
		best_cost_, counter_, double(ss-s)/CLOCKS_PER_SEC);
	
	return counter_;
}

double anneal_t::initial_temperature()
{
	double avg = 0;
	int uphill = 0;
	for (int j = 0; j < N_; ++j, ++counter_)
	{
		pt_->perturbation(-1);
		
		double cost = cost_->cost();

		if (cost > prev_cost_)
		{
			avg += prev_cost_-cost;
			++uphill;
		}
		else if (cost < best_cost_)
		{
			best_cost_ = cost;
			pt_->record_best();
			if (fp_ != 0) fprintf(fp_, "best cost %.3f\n", best_cost_);
		}
		prev_cost_ = cost;
	}
	avg /= uphill;

	// initial temperature
	double T = avg/log(0.85);
	if (fp_ != 0) fprintf(fp_, "uphill = %d, avg = %e, T = %f\n", uphill, avg, T);

	return T;
}

double anneal_t::round(double T, int n, double &avg)
{
	int MT = 0, uphill = 0, reject = 0;
	for (; (uphill < n) && (MT < 2*n); ++MT, ++counter_)
	{
		pt_->perturbation(T);
		double cost = cost_->cost();

		if ((cost < prev_cost_) || double(rand())/RAND_MAX < exp((prev_cost_-cost)/T))
		{
			if (prev_cost_ < cost)
			{
				avg += cost-prev_cost_;
				uphill++;
			}
			if (cost < best_cost_)
			{
				best_cost_ = cost;
				pt_->record_best();
				if (fp_ != 0) fprintf(fp_, "best cost %.3f\n", best_cost_);
			}
			prev_cost_ = cost;
		}
		else
		{
			reject++;
			pt_->reject();
		}
	}

	double rr = double(reject)/MT;
	if (fp_ != 0) fprintf(fp_,
		"uphill = %d, reject = %d, reject_rate = %f\n",
		uphill, reject, rr);

	return rr;
}

greedy_t::greedy_t(cost_ptr &cost, perturb_ptr &pt, int N, FILE *fp)
	: cost_(cost), pt_(pt), fp_(fp), N_(N), best_cost_(0), counter_(0)
{
}

double greedy_t::round(int n)
{
	int reject = 0;
	for (int i = 0; i < n; ++i, ++counter_)
	{

		pt_->perturbation(-1);
		double cost = cost_->cost();

		if (cost < best_cost_*0.99999999)
		{
			best_cost_ = cost;
			if (fp_ != 0) fprintf(fp_, "best cost %.3f\n", best_cost_);
		}
		else
		{
			++reject;
			pt_->reject();
		}
	}

	double rr = double(reject)/n;
	if (fp_ != 0) fprintf(fp_, "reject = %d, reject_rate = %f\n", reject, rr);

	return rr;
}

int greedy_t::go(double max_rr)
{
	clock_t s = clock();

	best_cost_ = cost_->cost();
	counter_ = 0;

	if (fp_ != 0) fprintf(fp_, "initial cost %.3f\n", best_cost_);

	// greedy search
	for (; round(N_) < max_rr;)
		;

	clock_t ss = clock();

	if (fp_ != 0) fprintf(fp_,
		"final %.3f, counter %d, time %f\n",
		best_cost_, counter_, double(ss-s)/CLOCKS_PER_SEC);
	
	return counter_;
}

int greedy_t::go_time(double sec)
{
	clock_t s = clock();

	best_cost_ = cost_->cost();
	counter_ = 0;

	if (fp_ != 0) fprintf(fp_, "initial cost %.3f\n", best_cost_);

	// greedy search
	for (; double(clock()-s)/CLOCKS_PER_SEC > sec;)
		 round(N_);

	clock_t ss = clock();

	if (fp_ != 0) fprintf(fp_,
		"final %.3f, counter %d, time %f\n",
		best_cost_, counter_, double(ss-s)/CLOCKS_PER_SEC);
	
	return counter_;
}

} // namespace detail

} // namespace fplan
