/*
 * Copyright (c) 2004-2005 Northwestern University.
 *                         NuCAD Group.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/*
 * cost.cpp - Implement cost expressions.
 *
 * Authors: Hai Zhou, haizhou@ece.northwestern.edu
 *          Jia Wang, jwa112@ece.northwestern.edu
 */

#include "../misc/common.h"

#include "cost.h"

namespace misc
{

using fplan::cost_ptr;

namespace detail
{

// although functors work here
// we don't want bother
typedef double (*func_t)(double, double);

class cost_binary_t : public fplan::detail::cost_i
{
	cost_ptr c_;
	cost_ptr cc_;
	func_t func_;

	virtual double cost() {return func_(c_->cost(), cc_->cost());}
	virtual void release() {delete this;}

	~cost_binary_t() {}

public:
	cost_binary_t(const cost_ptr &c, const cost_ptr &cc, func_t func)
		: c_(c), cc_(cc), func_(func)
		{}
}; // class cost_binary_t

double add(double a, double b) {return a+b;}
double sub(double a, double b) {return a-b;}
double mul(double a, double b) {return a*b;}
double div(double a, double b) {return a/b;}

class double_cost_t : public fplan::detail::cost_i
{
	double d_;
	
	virtual double cost() {return d_;}
	virtual void release() {delete this;}

	~double_cost_t() {}

public:
	double_cost_t(double d) : d_(d) {}
}; // class double_cost_t

class sqrt_cost_t : public fplan::detail::cost_i
{
	cost_ptr c_;
	
	virtual double cost() {return ::sqrt(c_->cost());}
	virtual void release() {delete this;}

	~sqrt_cost_t() {}

public:
	sqrt_cost_t(const cost_ptr &c) : c_(c) {}
}; // class sqrt_cost_t

} // namespace detail

cost_ptr double_cost(double d)
{
	return cost_ptr(new detail::double_cost_t(d));
}

cost_ptr sqrt(const cost_ptr &c)
{
	return cost_ptr(new detail::sqrt_cost_t(c));
}

cost_ptr operator + (const cost_ptr &c, const cost_ptr &cc)
{
	return cost_ptr(new detail::cost_binary_t(c, cc, detail::add));
}

cost_ptr operator - (const cost_ptr &c, const cost_ptr &cc)
{
	return cost_ptr(new detail::cost_binary_t(c, cc, detail::sub));
}

cost_ptr operator * (const cost_ptr &c, const cost_ptr &cc)
{
	return cost_ptr(new detail::cost_binary_t(c, cc, detail::mul));
}

cost_ptr operator / (const cost_ptr &c, const cost_ptr &cc)
{
	return cost_ptr(new detail::cost_binary_t(c, cc, detail::div));
}

extern "C" double pow(double, double);
cost_ptr pow(const cost_ptr &c, const cost_ptr &cc)
{
	return cost_ptr(new detail::cost_binary_t(c, cc, pow));
}

cost_ptr operator + (double d, const cost_ptr &c)
{
	return double_cost(d)+c;
}

cost_ptr operator + (const cost_ptr &c, double d)
{
	return c+double_cost(d);
}

cost_ptr operator - (double d, const cost_ptr &c)
{
	return double_cost(d)-c;
}

cost_ptr operator - (const cost_ptr &c, double d)
{
	return c-double_cost(d);
}

cost_ptr operator * (double d, const cost_ptr &c)
{
	return double_cost(d)*c;
}

cost_ptr operator * (const cost_ptr &c, double d)
{
	return c*double_cost(d);
}

cost_ptr operator / (double d, const cost_ptr &c)
{
	return double_cost(d)/c;
}

cost_ptr operator / (const cost_ptr &c, double d)
{
	return c/double_cost(d);
}

cost_ptr pow(double d, const cost_ptr &c)
{
	return pow(double_cost(d), c);
}

cost_ptr pow(const cost_ptr &c, double d)
{
	return pow(c, double_cost(d));
}

} // namespace misc
