/*
 * Copyright (c) 2004-2005 Northwestern University.
 *                         NuCAD Group.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/*
 * fplan_acg.cpp - Implement ACG fploorplanning.
 *
 * Authors: Hai Zhou, haizhou@ece.northwestern.edu
 *          Jia Wang, jwa112@ece.northwestern.edu
 *
 */

#include "../misc/common.h"

#include "fplan_acg.h"

#include "../acg/cg_pack.h"

namespace fplan
{

class acg_perturb_t : public detail::perturb_i
{
protected:
	acg_fp_t *fp_;
	double rotateP_, exchangeP_;

	acg::acg_base_t best_, temp_;
	std::vector<size_t> best_rotated_;

	int choice_;
	size_t n1_, n2_;

	virtual void perturbation(double T);
	virtual void reject();
	virtual void record_best();
	virtual void swap_best();

	// you should use the shared_ptr interface
	virtual void release() {delete this;}
	~acg_perturb_t() {}

public:
	acg_perturb_t(acg_fp_t *fp, double rotateP, double exchangeP);

}; // acg_perturb_t

acg_fp_t::acg_fp_t(size_t n, misc::dbl_pair shapes[], size_t ids[], size_t rotated[])
	: rcg_fp_t(n, shapes, ids, rotated, false), reduced_(true)
{
	acg::racg_build(acg_base_);
}

acg_fp_t::acg_fp_t(acg_fp_t &r)
	: reduced_(true)
{
	r.require_racg();

	acg_base_ = r.acg_base_;
	rotated_ = r.rotated_;
	shape_ = r.shape_;
	mapped_ = r.mapped_;
}

acg_fp_t::~acg_fp_t()
{
}

void acg_fp_t::swap(acg_fp_t &r)
{
	acg_base_.swap(r.acg_base_);
	rotated_.swap(r.rotated_);
	std::swap(shape_, r.shape_);
	std::swap(mapped_, r.mapped_);

	ctx_.swap(r.ctx_);
	std::swap(reduced_, r.reduced_);
}

acg_fp_t &acg_fp_t::operator = (acg_fp_t r)
{
	swap(r);

	return *this;
}

void acg_fp_t::require_acg()
{
	if (reduced_)
	{
		acg::racg_to_acg(acg_base_, ctx_);
		reduced_ = false;
	}
}

void acg_fp_t::require_racg()
{
	if (!reduced_)
	{
		acg::acg_to_racg(acg_base_, ctx_);
		reduced_ = true;
	}
}

void acg_fp_t::do_packing()
{
	if (!mapped_)
	{
		require_acg();

		shape_ = acg::packing_longest_path(acg_base_);

		mapped_ = true;
	}
}

void acg_fp_t::compute_bb()
{
	do_packing();

	require_acg();

	acg::longest_path_bb(acg_base_, shape_);
}

perturb_ptr acg_fp_t::create_perturbation(double rotateP, double exchangeP)
{
	return perturb_ptr(new acg_perturb_t(this, rotateP, exchangeP));
}

acg_perturb_t::acg_perturb_t(acg_fp_t *fp, double rotateP, double exchangeP)
	: fp_(fp),
	rotateP_(rotateP), exchangeP_(exchangeP),
	choice_(-1), n1_(0), n2_(0)
{
}

void acg_perturb_t::perturbation(double T)
{
	const size_t n = fp_->acg_base_.num_nodes();

	assert(n != 0);

	double P = double(rand())/RAND_MAX;

	fp_->mapped_ = false;

	if ((P < rotateP_) || (n == 1))
	{
		choice_ = 0;
		
		n1_ = rand()%n;

		fp_->acg_base_.node_rotate(n1_);
		fp_->rotated_[n1_] = 1-fp_->rotated_[n1_];
	}
	else if (P < rotateP_+exchangeP_)
	{
		choice_ = 1;
	
		n1_ = rand()%fp_->acg_base_.num_nodes();
	
		do
		{
			n2_ = (int)rand()%n;
		}
		while (n2_ == n1_);

		fp_->acg_base_.node_swap(n1_, n2_);
		std::swap(fp_->rotated_[n1_], fp_->rotated_[n2_]);
	}
	else
	{
		choice_ = 2;

		P -= rotateP_+exchangeP_;
		double ref = 1-rotateP_-exchangeP_;

		fp_->require_racg();
		temp_ = fp_->acg_base_; // force racg and copy
	
		acg::pos_t pos = (rand()%2 == 0)? acg::POS_IN: acg::POS_OUT;
		if (P < 0.34*ref)
			acg::racg_switch_adjacent(fp_->acg_base_, n1_, pos);
		else if (P < 0.67*ref)
			acg::racg_switch_alter(fp_->acg_base_, n1_, pos);
		else
			acg::racg_switch_last(fp_->acg_base_, n1_, pos);
	}
}

void acg_perturb_t::reject()
{
	fp_->mapped_ = false;

	switch (choice_)
	{
	case 0:
		fp_->acg_base_.node_rotate(n1_);
		fp_->rotated_[n1_] = 1-fp_->rotated_[n1_];
		break;
	case 1:
		fp_->acg_base_.node_swap(n1_, n2_);
		std::swap(fp_->rotated_[n1_], fp_->rotated_[n2_]);
		break;
	case 2:
		fp_->acg_base_.swap(temp_);
		fp_->reduced_ = true;
		fp_->ctx_.clear();
		break;
	default:
		assert(false);
	}
}

void acg_perturb_t::record_best()
{
	fp_->require_racg();

	best_ = fp_->acg_base_;
	best_rotated_ = fp_->rotated_;
}

void acg_perturb_t::swap_best()
{
	best_.swap(fp_->acg_base_);
	best_rotated_.swap(fp_->rotated_);
	fp_->mapped_ = false;

	fp_->ctx_.clear();
	fp_->reduced_ = true;
}

} // namespace fplan
