/*
 * Copyright (c) 2004-2005 Northwestern University.
 *                         NuCAD Group.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/*
 * nets.cpp - Nets/hyper-graph interface.
 *
 * Authors: Hai Zhou, haizhou@ece.northwestern.edu
 *          Jia Wang, jwa112@ece.northwestern.edu
 *
 */

#include "../misc/common.h"

#include "../misc/linear_eqn.h"

#include "nets.h"

namespace fplan
{

module_nets_t::module_nets_t()
{
	nets_.push_back(0);
}

void module_nets_t::append_net()
{
	nets_.push_back(net_modules_.size());
}

void module_nets_t::append_module(size_t im)
{
	net_modules_.push_back(im);
	++nets_.back();
}

double module_nets_t::HPWL(const misc::dbl_pair centers[]) const
{
	double wl = 0;

	for (size_t i = 0; i < nets_.size()-1; ++i)
	{
		size_t from = nets_[i];
		size_t to = nets_[i+1];

		if (from+2 == to) // two pin nets
		{
			const misc::dbl_pair pt0 = centers[net_modules_[from]];
			const misc::dbl_pair pt1 = centers[net_modules_[from+1]];
			wl += fabs(pt0.first-pt1.first)+fabs(pt0.second-pt1.second);
			continue;
		}
		else if (from+3 == to) // three pin nets
		{
			const misc::dbl_pair pt0 = centers[net_modules_[from]];
			const misc::dbl_pair pt1 = centers[net_modules_[from+1]];
			const misc::dbl_pair pt2 = centers[net_modules_[from+2]];
			const misc::dbl_pair ll = misc::min(pt0, misc::min(pt1, pt2));
			const misc::dbl_pair ur = misc::max(pt0, misc::max(pt1, pt2));
			wl += ur.first-ll.first+ur.second-ll.second;
			continue;
		}
		else if (from+4 == to) // four pin nets
		{
			const misc::dbl_pair pt0 = centers[net_modules_[from]];
			const misc::dbl_pair pt1 = centers[net_modules_[from+1]];
			const misc::dbl_pair pt2 = centers[net_modules_[from+2]];
			const misc::dbl_pair pt3 = centers[net_modules_[from+2]];
			const misc::dbl_pair ll = misc::min(misc::min(pt0, pt1), misc::min(pt2, pt3));
			const misc::dbl_pair ur = misc::max(misc::max(pt0, pt1), misc::max(pt2, pt3));
			wl += ur.first-ll.first+ur.second-ll.second;
			continue;
		}
		else if ((from == to) || (from+1 == to))
			continue;

		misc::dbl_pair ll = centers[net_modules_[from]];
		misc::dbl_pair ur = ll;

		for (++from; from != to; ++from)
		{
			const misc::dbl_pair pt = centers[net_modules_[from]];
			
			ll = misc::min(pt, ll);
			ur = misc::max(pt, ur);
		}

		wl += ur.first-ll.first+ur.second-ll.second;
	}

	return wl;
}

double module_nets_t::QPWL(const misc::dbl_pair centers[]) const
{
	double wl = 0;

	for (size_t i = 0; i < nets_.size()-1; ++i)
	{
		size_t from = nets_[i];
		size_t to = nets_[i+1];

		if ((from == to) || (from+1 == to))
			continue;

		misc::dbl_pair pt;
		for (size_t j = from; j != to; ++j)
			pt += centers[net_modules_[j]];
		pt /= to-from;

		double ql = 0;
		for (size_t k = from; k != to; ++k)
		{
			misc::dbl_pair d = centers[net_modules_[k]]-pt;
			ql += d.first*d.first+d.second*d.second;
		}

		wl += ql/(to-from);
	}

	return wl;
}

bool module_nets_t::QP(size_t n, misc::dbl_pair locs[], const misc::dbl_pair pins[]) const
{
	using linear::v_double;
	std::vector<v_double> Q(n, v_double(n, 0));
	v_double x(n, 0), y(n, 0);

	// stamping
	for (size_t i = 0; i < nets_.size()-1; ++i)
	{
		size_t from = nets_[i];
		size_t to = nets_[i+1];

		if ((from == to) || (from+1 == to))
			continue;

		// 1/(k-1) for every edge in a k-clique
		double weight = 1.0/(to-from-1);

		for (size_t j = from; j != to; ++j)
			for (size_t k = j+1; k != to; ++k)
			{
				size_t m0 = net_modules_[j];
				size_t m1 = net_modules_[k];

				if (m0 > m1)
					std::swap(m0, m1);
				else if (m0 == m1)
					continue;

				// now m0 < m1
				if (n <= m0)
					// both are pins
					continue;

				if (m1 < n)
					// both are modules
				{
					Q[m0][m0] += weight;
					Q[m1][m1] += weight;
					Q[m0][m1] += -weight;
					Q[m1][m0] += -weight;
				}
				else
					// m0 is module, m1 is pin
				{
					Q[m0][m0] += 5*weight;
					x[m0] += 5*weight*pins[m1-n].first;
					y[m0] += 5*weight*pins[m1-n].second;
				}
			}
	}

	if (!linear::Gauss2(n, &Q[0], &x[0], &y[0]))
		return false;

	for (size_t i = 0; i < n; ++i)
		locs[i] = misc::dbl_pair(x[i], y[i]);

	return true;
}

pin_nets_t::pin_nets_t()
{
	nets_.push_back(0);
}

void pin_nets_t::append_net()
{
	nets_.push_back(net_pins_.size());
}

void pin_nets_t::append_pin(size_t ip, const misc::dbl_pair &loc)
{
	net_pins_.push_back(std::make_pair(ip, loc));
	++nets_.back();
}

double pin_nets_t::HPWL(size_t nm,
	const misc::dbl_pair shapes[],
	const size_t rotations[],
	const misc::dbl_pair centers[]) const
{
	double wl = 0;

	for (size_t i = 0; i < nets_.size()-1; ++i)
	{
		size_t from = nets_[i];
		size_t to = nets_[i+1];

		if ((from == to) || (from+1 == to))
			continue;

		size_t old_from = from;

		misc::dbl_pair ll, ur;

		for (; from != to; ++from)
		{
			const pin_t &pin = net_pins_[from];
			misc::dbl_pair pt = centers[pin.first];
			const misc::dbl_pair &s = shapes[pin.first];
			const misc::dbl_pair &p = pin.second;
			
			if (pin.first < nm)
				switch (rotations[pin.first])
				{
				case 0:
					pt.first += s.first*p.first;
					pt.second += s.second*p.second;
					break;
				case 1:
					pt.first += s.second*p.second;
					pt.second -= s.first*p.first;
					break;
				case 2:
					pt.first -= s.first*p.first;
					pt.second -= s.second*p.second;
					break;
				case 3:
					pt.first -= s.second*p.second;
					pt.second += s.first*p.first;
					break;
				case 4:
					pt.first += s.first*p.first;
					pt.second -= s.second*p.second;
					break;
				case 5:
					pt.first += s.second*p.second;
					pt.second += s.first*p.first;
					break;
				case 6:
					pt.first -= s.first*p.first;
					pt.second += s.second*p.second;
					break;
				case 7:
					pt.first -= s.second*p.second;
					pt.second -= s.first*p.first;
					break;
				default:
					assert(false);
				}

			if (from == old_from)
				ll = ur = pt;
			else
			{
				ll = misc::min(pt, ll);
				ur = misc::max(pt, ur);
			}
		}

		wl += ur.first-ll.first+ur.second-ll.second;
	}

	return wl;
}

} // namespace fplan
