/*
 * Copyright (c) 2004-2005 Northwestern University.
 *                         NuCAD Group.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/*
 * nets.h - Nets/hyper-graph interface.
 *
 * Authors: Hai Zhou, haizhou@ece.northwestern.edu
 *          Jia Wang, jwa112@ece.northwestern.edu
 *
 */

#ifndef NETS_H
#define NETS_H

#include "../misc/geometry.h"

namespace fplan
{

// nets that contain only modules
class module_nets_t
{
	std::vector<size_t> net_modules_;
	std::vector<size_t> nets_;

public:
	module_nets_t();

	size_t num_nets() const {return nets_.size()-1;}
	size_t num_modules(size_t inet) const
	{
		assert(inet < num_nets());
		return nets_[inet+1]-nets_[inet];
	}
	size_t module(size_t inet, size_t im) const
	{
		assert(inet < num_nets());
		assert(im < num_modules(inet));
		return net_modules_[nets_[inet]+im];
	}

	// add nets one by one
	void append_net();
	void append_module(size_t im);

	// QP
	bool QP(size_t n, misc::dbl_pair locs[], const misc::dbl_pair pins[]) const;

	// WL functions
	double HPWL(const misc::dbl_pair centers[]) const;
	double QPWL(const misc::dbl_pair centers[]) const;
}; // class module_nets_t

// nets that contain pins
class pin_nets_t
{
public:
	typedef std::pair<size_t, misc::dbl_pair> pin_t;

private:
	std::vector<pin_t> net_pins_;
	std::vector<size_t> nets_;

public:
	pin_nets_t();

	size_t num_nets() const {return nets_.size()-1;}
	size_t num_pins(size_t inet) const
	{
		assert(inet < num_nets());
		return nets_[inet+1]-nets_[inet];
	}
	pin_t pin(size_t inet, size_t ip) const
	{
		assert(inet < num_nets());
		assert(ip < num_pins(inet));
		return net_pins_[nets_[inet]+ip];
	}

	// add nets one by one
	void append_net();
	void append_pin(size_t ip, const misc::dbl_pair &loc);

	// WL functions
	double HPWL(size_t nm, const misc::dbl_pair shapes[],
		const size_t rotations[], const misc::dbl_pair centers[]) const;

/*
	typedef double (pin_nets_t::*WL_func_t)(
		size_t nm, const misc::dbl_pair shapes[],
		const size_t rotations[],
		const misc::dbl_pair centers[]) const;

	// cost expression evaluation is possible
	template <class T, class CentersFunc>
	cost_ptr cost(T t, CentersFunc func, WL_func_t wl_func,
		size_t nm, const misc::dbl_pair shapes[], const size_t rotations[],
		size_t np, const misc::dbl_pair pins[])
	{
		struct wl_t : public detail::cost_i
		{
			T t_;
			CentersFunc func_;
			pin_nets_t &pnets_;
			WL_func_t wl_func_;
			size_t nm_, np_;
			const misc::dbl_pair *shapes_;
			const size_t *rotations_;
			const misc::dbl_pair *pins_;
			std::vector<misc::dbl_pair> centers_;

			virtual double cost()
			{
				// get shape and centers
				misc::dbl_pair shape = func_(t_, &centers_[0]);

				// compute pin positions
				for (size_t i = 0; i < np_; ++i)
					centers_[i+nm_] = pins_[i]*shape;

//				for (size_t j = 0; j < nm_+np_; ++j)
//					printf("%u, %.0f, %.0f\n", j, centers_[j].first, centers_[j].second);

				return (pnets_.*wl_func_)(nm_, shapes_, rotations_, &centers_[0]);
			}

			virtual void release() {delete this;}

			wl_t(const T &t, CentersFunc func,
				pin_nets_t &pnets, WL_func_t wl_func,
				size_t nm, const misc::dbl_pair shapes[], const size_t rotations[],
				size_t np, const misc::dbl_pair pins[])
				: t_(t), func_(func), pnets_(pnets), wl_func_(wl_func),
				nm_(nm), np_(np), shapes_(shapes),
				rotations_(rotations), pins_(pins), centers_(nm+np)
				{}

			// mingw 3.4.2 gcc won't work without this
			~wl_t() {}
		}; // struct area_t

		return cost_ptr(new wl_t(t, func, *this, wl_func, nm, shapes, rotations, np, pins));
	}

	// shortcuts
	template <class T, class CentersFunc>
	cost_ptr HPWL_cost(T t, CentersFunc func,
		size_t nm, const misc::dbl_pair shapes[], const size_t rotations[],
		size_t np, const misc::dbl_pair pins[])
		{return cost(t, func, &pin_nets_t::HPWL, nm, shapes, rotations, np, pins);}
*/
}; // class pin_nets_t

} // namespace fplan

#endif // NETS_H
