/*
 * Copyright (c) 2004-2005 Northwestern University.
 *                         NuCAD Group.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/*
 * fplan_demo.cpp - Just a demo.
 *
 * Authors: Hai Zhou, haizhou@ece.northwestern.edu
 *          Jia Wang, jwa112@ece.northwestern.edu
 *
 */

#include "misc/common.h"

#include "gsrc/gsrc_info.h"
#include "fplan/fplan.h"
#include "fplan/fplan_acg.h"
#include "fplan/fplan_rcg.h"

using namespace fplan;

std::string g_bench;

std::vector<gsrc::module_t> g_modules;
std::vector<gsrc::signal_t> g_signals;

std::vector<misc::dbl_pair> g_module_shapes;
std::vector<misc::cstr_t> g_module_names;
std::vector<misc::dbl_pair> g_pin_locations;
module_nets_t g_mnets;

double g_total = 0;

void SA_optimize(fplan_i *fp, perturb_ptr perturb, bool wire);

int usage(const char *exe)
{
	fprintf(stderr, "usage:\n"
		"%s acg|rcg bench_name [wire]\n",
		exe);

	return 0;
}

std::string filename(const std::string &pathname)
{
	std::string ret;
	for (size_t i = pathname.length(); i != 0; --i)
		if ((pathname[i-1] == '\\') || (pathname[i-1] == '/'))
			break;
		else
			ret = std::string(1, pathname[i-1])+ret;
	return ret;
}

bool load_gsrc()
{
	int n = gsrc::read_gsrc(g_bench, g_modules, g_signals);

	if (n <= 0)
	{
		fprintf(stderr, "Cannot read GSRC benchmark %s\n", g_bench.c_str());
		return false;
	}

	g_bench = filename(g_bench);

	if (!fplan::gsrc_aux(g_bench+"_aux.xml", n, g_modules, g_signals))
		fprintf(stderr, "Cannot write aux file %s_aux.xml\n", g_bench.c_str());

	gsrc_info(n,
		g_modules, g_signals,
		g_module_shapes, g_module_names,
		g_pin_locations,
		g_mnets);

	g_total = 0;
	for (size_t i = 0; i < g_module_shapes.size(); ++i)
		g_total += g_module_shapes[i].first*g_module_shapes[i].second;

	fprintf(stdout, "total %e\n", g_total);

	return true;
}

int main(int argc, char *argv[])
{
	bool use_acg = true;

	if ((argc >= 3) && (strcmp(argv[1], "acg") == 0))
		;
	else if ((argc >= 3) && (strcmp(argv[1], "rcg") == 0))
		use_acg = false;
	else
		return usage(argv[0]);

	g_bench = argv[2];
	bool wire = (argc > 3);

	if (!load_gsrc())
		return 0;

	srand(0);

	if (use_acg)
	{
		acg_fp_t acg_fp(g_module_shapes.size(), &g_module_shapes[0]);

		SA_optimize(acg_fp.get_fplan_interface(),
			acg_fp.create_perturbation(0.33, 0.33), wire);

		acg_fp.dump(g_bench+"_acg.xml", &g_module_names[0]);
	}
	else
	{
		rcg_fp_t rcg_fp(g_module_shapes.size(), &g_module_shapes[0]);

		SA_optimize(rcg_fp.get_fplan_interface(),
			rcg_fp.create_perturbation(0.33, 0.33), wire);

		rcg_fp.dump(g_bench+"_rcg.xml", &g_module_names[0]);
	}

	return 0;
}

void SA_optimize(fplan_i *fp, perturb_ptr perturb, bool wire)
{
	const size_t n = g_module_shapes.size();
	const size_t m = g_pin_locations.size();
	std::string oname;

	cost_ptr wlHP = fp->HPWL_cost(g_mnets, m, &g_pin_locations[0]);
	cost_ptr wlQP = fp->QPWL_cost(g_mnets, m, &g_pin_locations[0]);
	cost_ptr area = fp->area_cost();
	cost_ptr square = fp->square_cost();

	cost_ptr cost = area;
	if (wire)
	{
		if (n < 50)
			cost = area/1000+wlHP; // normalize MCNC's unit
		else
			cost = area+wlHP;
	}

	fprintf(stdout,
		"area %e, total %e, square %e, "
		"HPWL %e, QPWL %e, target %e\n",
		area->cost(), g_total, square->cost(),
		wlHP->cost(), wlQP->cost(), cost->cost());

	anneal(cost, perturb, int(n), stdout);

	fprintf(stdout,
		"area %e, total %e, square %e, "
		"HPWL %e, QPWL %e, target %e\n",
		area->cost(), g_total, square->cost(),
		wlHP->cost(), wlQP->cost(), cost->cost());
}
