/*
 * Copyright (c) 2004-2005 Northwestern University.
 *                         NuCAD Group.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/*
 * gsrc.cpp - Read GSRC benchmark.
 *
 * Authors: Jia Wang, jwa112@ece.northwestern.edu
 *
 */

#include "../misc/common.h"

#include "../misc/token_parser.h"

#include "gsrc.h"

namespace gsrc
{

using namespace misc;

typedef std::map<std::string, module_t *> map_modules;

int gsrc_blocks(const std::string &blocks,
	std::vector<module_t> &modules, map_modules &names);

bool gsrc_nets(const std::string &nets,
	std::vector<signal_t> &signals, map_modules &names,
	std::vector<detailed_signal_t> *detailed);

bool gsrc_pl(const std::string &pl,
	std::vector<module_t> &modules, map_modules &names);

int read_gsrc(const std::string &gsrc,
	std::vector<module_t> &modules,
	std::vector<signal_t> &signals,
	std::vector<detailed_signal_t> *detailed,
	misc::dbl_pair *shape)
{
	assert(modules.size() == 0);
	assert(signals.size() == 0);

	std::string blocks = gsrc+".blocks";
	std::string nets = gsrc+".nets";
	std::string pl = gsrc+".pl";
	std::string pl2 = gsrc+"_I.pl"; // for MCNC

	map_modules names;

	int nblocks = gsrc_blocks(blocks, modules, names);

	if (nblocks == 0)
	{
		modules.clear();
		signals.clear();
		return 0;
	}

	if (!gsrc_nets(nets, signals, names, detailed))
		signals.clear();

	if (!gsrc_pl(pl, modules, names))
		if (!gsrc_pl(pl2, modules, names))
			return nblocks;

	// calculating terminal positions
	double maxx = -1e9, minx = 1e9, maxy = -1e9, miny = 1e9;
	for (size_t i = nblocks; i < modules.size(); ++i)
	{
		if (modules[i].x > maxx) maxx = modules[i].x;
		if (modules[i].y > maxy) maxy = modules[i].y;
		if (modules[i].x < minx) minx = modules[i].x;
		if (modules[i].y < miny) miny = modules[i].y;
	}
	if (maxx > minx)
		for (size_t i = nblocks; i < modules.size(); ++i)
			modules[i].w = (modules[i].x-minx)/(maxx-minx);
	if (maxy > miny)
		for (size_t i = nblocks; i < modules.size(); ++i)
			modules[i].h = (modules[i].y-miny)/(maxy-miny);

	if (shape != 0)
	{
		shape->first = maxx-minx;
		shape->second = maxy-miny;
	}

	return nblocks;
}

const cstr_t gsrc_delimit = " \r\n\t,()=:";

int gsrc_blocks(const std::string &blocks,
	std::vector<module_t> &modules, map_modules &names)
{
	FILE *fp = fopen(blocks.c_str(), "r");
	if (fp == 0)
	{
		fprintf(stderr, "%s: no blocks file\n", blocks.c_str());
		return 0;
	}

	int nblocks = -1, nterminals = -1, iblock = 0, iterminal = 0;
	bool init = false;

	char linebuf[1000];
	size_t line = 0;
	while (fgets(linebuf, sizeof(linebuf), fp) != 0)
	{
		line++;
		char *temp = skip_space(linebuf, gsrc_delimit);
		if ((*temp == 0) || (*temp == '#'))
			continue; // empty line and comments

		std::vector<cstr_t> tokens;
		split_tokens(temp, &tokens, gsrc_delimit);

		if (!init && (nblocks > 0) && (nterminals >= 0))
		{
			init = true;
			modules.resize(nblocks+nterminals);
			iblock = 0;
			iterminal = nblocks;
		}

		if (!init && (tokens.size() == 2)
			&& (strcmp(tokens[0], "NumHardRectilinearBlocks") == 0))
			nblocks = atoi(tokens[1]);
		else if (!init && (tokens.size() == 2)
			&& (strcmp(tokens[0], "NumTerminals") == 0))
			nterminals = atoi(tokens[1]);
		else if (init && (tokens.size() == 11)
			&& (strcmp(tokens[1], "hardrectilinear") == 0)
			&& (strcmp(tokens[2], "4") == 0))
		{
			if (iblock == nblocks)
			{
				fprintf(stderr, "%s:%d: error: extra hard block\n", blocks.c_str(), line);
				fclose(fp);
				return 0;
			}
			
			module_t &module = modules[iblock];
			module.name = tokens[0];
			module.w = atof(tokens[7]);
			module.h = atof(tokens[8]);
			module.x = module.y = 0;
			module.id = 0;

			names[tokens[0]] = &module;
			iblock++;
		}
		else if (init && (tokens.size() == 2)
			&& (strcmp(tokens[1], "terminal") == 0))
		{
			if (iterminal == (int)modules.size())
			{
				fprintf(stderr, "%s:%d: error: extra terminal\n", blocks.c_str(), line);
				fclose(fp);
				return 0;
			}

			module_t &module = modules[iterminal];
			module.name = tokens[0];
			module.w = module.h = 0;
			module.x = module.y = 0;
			module.id = -1;

			names[tokens[0]] = &module;
			iterminal++;
		}
	}

	if (iblock != nblocks)
	{
		fprintf(stderr,
			"%s: error: missing %d hard blocks\n",
			blocks.c_str(), nblocks-iblock);
		fclose(fp);
		return 0;
	}
			
	if (iterminal != (int)modules.size())
	{
		fprintf(stderr,
			"%s: error: missing %d terminals\n",
			blocks.c_str(), modules.size()-iterminal);
		fclose(fp);
		return 0;
	}

	fclose(fp);
	return nblocks;
}

bool gsrc_nets(const std::string &nets,
	std::vector<signal_t> &signals, map_modules &names,
	std::vector<detailed_signal_t> *detailed)
{
	FILE *fp = fopen(nets.c_str(), "r");
	if (fp == 0)
	{
		fprintf(stderr, "%s: no nets file\n", nets.c_str());
		return false;
	}

	int nnets = -1, npins = -1, inet = 0, ipin = 0, degree = 0;
	bool init = false;

	char linebuf[1000];
	size_t line = 0;
	while (fgets(linebuf, sizeof(linebuf), fp) != 0)
	{
		line++;
		char *temp = skip_space(linebuf, gsrc_delimit);
		if ((*temp == 0) || (*temp == '#'))
			continue; // empty line and comments

		std::vector<cstr_t> tokens;
		split_tokens(temp, &tokens, gsrc_delimit);

		if (!init && (nnets > 0) && (npins > 0))
		{
			init = true;
			signals.resize(nnets);
			if (detailed != 0)
				detailed->resize(nnets);
			inet = -1;
			ipin = 0;
		}

		if (!init && (tokens.size() == 2)
			&& (strcmp(tokens[0], "NumNets") == 0))
			nnets = atoi(tokens[1]);
		else if (!init && (tokens.size() == 2)
			&& (strcmp(tokens[0], "NumPins") == 0))
			npins = atoi(tokens[1]);
		else if (init && (tokens.size() == 2)
			&& (strcmp(tokens[0], "NetDegree") == 0))
		{
			if (degree != 0)
			{
				fprintf(stderr,
					"%s:%d: error: missing %d pins",
					nets.c_str(), line, degree);
				fclose(fp);
				return false;
			}
			degree = atoi(tokens[1]);
			++inet;
			if (inet == nnets)
			{
				fprintf(stderr, "%s:%d: error: extra nets\n", nets.c_str(), line);
				fclose(fp);
				return false;
			}
		}
		else if (init && (degree > 0)
			&& (tokens.size() >= 2))
		{
			map_modules::iterator it = names.find(tokens[0]);
			if (it == names.end())
			{
				fprintf(stderr, "%s:%d: error: unknown pin %s\n",
					nets.c_str(), line, tokens[0]);
				fclose(fp);
				return false;
			}
			signals[inet].pins.push_back(it->second);
			it->second->nets.push_back(&signals[inet]);

			if (detailed != 0) // add detailed pin info
			{
				detailed_signal_t::pin_t pin;
				pin.first = it->second;

				if (tokens.size() == 4)
				{
					if ((tokens[2][0] != '%') || (tokens[3][0] != '%'))
					{
						fprintf(stderr, "%s:%d: error: pin position %s\n",
							nets.c_str(), line, tokens[0]);
						fclose(fp);
						return false;
					}
					pin.second.first = atof(tokens[2]+1)/100;
					pin.second.second = atof(tokens[3]+1)/100;
//					printf("%s: %.2f, %.2f\n", tokens[0], pin.second.first, pin.second.second);
				}
				else if (tokens.size() != 2)
				{
					fprintf(stderr, "%s:%d: error: pin %s\n",
						nets.c_str(), line, tokens[0]);
					fclose(fp);
					return false;
				}

				detailed->at(inet).pins.push_back(pin);
			}
			
			--degree;
			++ipin;
		}
	}

	++inet;
	if (inet != nnets)
	{
		fprintf(stderr,
			"%s: error: missing %d nets\n",
			nets.c_str(), nnets-inet);
		fclose(fp);
		return false;
	}
			
	if (ipin != npins)
	{
		fprintf(stderr,
			"%s: error: missing %d pins\n",
			nets.c_str(), npins-ipin);
		fclose(fp);
		return false;
	}

	fclose(fp);
	return true;
}

bool gsrc_pl(const std::string &pl,
	std::vector<module_t> &modules, map_modules &names)
{
	FILE *fp = fopen(pl.c_str(), "r");
	if (fp == 0)
	{
		fprintf(stderr, "%s: no pl file\n", pl.c_str());
		return false;
	}

	bool init = false;

	char linebuf[1000];
	size_t line = 0;
	while (fgets(linebuf, sizeof(linebuf), fp) != 0)
	{
		line++;
		char *temp = skip_space(linebuf, gsrc_delimit);
		if ((*temp == 0) || (*temp == '#'))
			continue; // empty line and comments

		std::vector<cstr_t> tokens;
		split_tokens(temp, &tokens, gsrc_delimit);

		if (!init && (tokens.size() == 3)
			&& ((strcmp(tokens[0], "UCSC") == 0) || (strcmp(tokens[0], "UCLA") == 0)))
			init = true;
		else if (init && (tokens.size() == 3))
		{
			map_modules::iterator it = names.find(tokens[0]);
			if (it == names.end())
			{
				fprintf(stderr, "%s:%d: warning: unknown pin %s\n",
					pl.c_str(), line, tokens[0]);
				fclose(fp);
				return false;
			}
			it->second->x = atof(tokens[1]);
			it->second->y = atof(tokens[2]);
		}
	}

	fclose(fp);
	return true;
}

} // namespace gsrc
