/*
 * Copyright (c) 2004-2005 Northwestern University.
 *                         NuCAD Group.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/*
 * gsrc_info.cpp - Get infomation from GSRC benchmark to be used in fplan.
 *
 * Authors: Jia Wang, jwa112@ece.northwestern.edu
 *
 */

#include "../misc/common.h"

#include "gsrc_info.h"

namespace fplan
{

void gsrc_info(int n,
	const std::vector<gsrc::module_t> &modules,
	const std::vector<gsrc::signal_t> &signals,
	std::vector<misc::dbl_pair> &module_shapes,
	std::vector<misc::cstr_t> &module_names,
	std::vector<misc::dbl_pair> &pin_locations,
	module_nets_t &mnets,
	const std::vector<gsrc::detailed_signal_t> *detailed,
	pin_nets_t *pnets)
{
	const size_t N = size_t(n);

	assert(modules.size() >= N);

	module_shapes.reserve(N);
	module_names.reserve(N);
	pin_locations.reserve(modules.size()-N);

	for (size_t i = 0; i < N; ++i)
	{
		module_shapes.push_back(misc::dbl_pair(modules[i].w, modules[i].h));
		module_names.push_back(modules[i].name.c_str());
	}
	for (size_t j = N; j < modules.size(); ++j)
		pin_locations.push_back(misc::dbl_pair(modules[j].w, modules[j].h));
	for (size_t k = 0; k < signals.size(); ++k)
	{
		mnets.append_net();
		for (size_t i = 0; i < signals[k].pins.size(); ++i)
			mnets.append_module(signals[k].pins[i]-&modules[0]);
	}
	if (detailed != 0)
	{
		for (size_t i = 0 ; i < detailed->size(); ++i)
		{
			pnets->append_net();
			for (size_t j = 0; j < detailed->at(i).pins.size(); ++j)
				pnets->append_pin(
					detailed->at(i).pins[j].first-&modules[0], 
					detailed->at(i).pins[j].second);
		}
	}
}

bool gsrc_aux(const std::string &aux_name, int n,
	const std::vector<gsrc::module_t> &modules,
	const std::vector<gsrc::signal_t> &signals)
{
	FILE *fp = fopen(aux_name.c_str(), "w");

	if (fp == 0)
		return false;

	fprintf(fp, "<?xml version=\"1.0\"?>\n");
	fprintf(fp, "<design>\n");

	fprintf(fp, "  <io_pins>\n");
	for (size_t i = size_t(n); i < modules.size(); ++i)
		fprintf(fp, "    <io_pin name='%s' x='%e' y='%e' />\n",
			modules[i].name.c_str(), modules[i].w, modules[i].h);
	fprintf(fp, "  </io_pins>\n");

	fprintf(fp, "  <nets>\n");
	for (size_t i = 0; i < signals.size(); ++i)
	{
		const gsrc::signal_t &s = signals[i];

		fprintf(fp, "    <net pins='");
		for (size_t j = 0; j < s.pins.size(); ++j)
			fprintf(fp, "%s,", s.pins[j]->name.c_str());
		fprintf(fp, "' />\n");
	}
	fprintf(fp, "  </nets>\n");

	fprintf(fp, "</design>\n");

	fclose(fp);

	return true;
}

} // namespace fplan
