# Copyright (c) 2004-2005 Northwestern University.
#                         NuCAD Group.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# Floorplan Viewer
#
# Authors: Jia Wang, jwa112@ece.northwestern.edu
#

import os
import wx
import xml.parsers.expat

g_title = 'Floorplan Viewer'
g_version = '0.1'
g_author = 'Jia Wang'

def Rect4ToRectWH(*rc):
    return ((rc[0], rc[1]), (rc[2]-rc[0], rc[3]-rc[1]))

def RectWHToRect4(pos, shape):
    return (pos[0], pos[1], pos[0]+shape[0], pos[1]+shape[1])

def CenterRect4(*rc):
    return (float(rc[0]+rc[2])/2, float(rc[1]+rc[3])/2)

def CenterRectWH(pos, shape):
    return (pos[0]+float(shape[0])/2, pos[1]+float(shape[1])/2)

def AttrsToBB(attrs):
    return (float(attrs[u'llx']), float(attrs[u'lly']), float(attrs[u'urx']), float(attrs[u'ury']))

def AttrsToShape(attrs):
    return (float(attrs[u'width']), float(attrs[u'height']))

def AttrsToPos(attrs):
    return (float(attrs[u'x']), float(attrs[u'y']))

class Module:

    def __init__(self, attrs):
        self.pos = AttrsToPos(attrs)
        self.shape = AttrsToShape(attrs)
        try:
            self.bb = AttrsToBB(attrs)
        except KeyError:
            self.bb = (self.pos[0], self.pos[1], self.pos[0]+self.shape[0], self.pos[1]+self.shape[1])

    def __repr__(self):
        return '%s, %s, %s' % (self.pos, self.shape, self.bb)

    def PosRect4(self):
        return RectWHToRect4(self.pos, self.shape)

    def BBRect4(self):
        return self.bb

    def BBCenterRect4(self):
        cx, cy = CenterRect4(*self.bb)
        return (cx-self.shape[0]/2, cy-self.shape[1]/2, cx+self.shape[0]/2, cy+self.shape[1]/2)

class Edge:

    def __init__(self, attrs):
        self.color = attrs[u'color']
        self.from_module = attrs[u'from']
        self.to_module = attrs[u'to']

    def __repr__(self):
        return '(%s->%s, %s)' % (self.from_module, self.to_module, self.color)

class IOPin:

    def __init__(self, attrs, pos, shape):
        rx, ry = AttrsToPos(attrs)
        self.pos = (rx*shape[0]+pos[0], ry*shape[1]+pos[1])

class Net:

    def __init__(self, attrs):
        self.pins = attrs[u'pins'].split(',')[:-1]

class Group:

    def __init__(self, attrs):
        self.modules = attrs[u'modules'].split(',')[:-1]

        try:
            self.bb = AttrsToBB(attrs)
        except KeyError:
            self.bb = (0.0, 0.0, 0.0, 0.0)


class Design:

    def __init__(self, attrs):
        self.pos = AttrsToPos(attrs)
        self.shape = AttrsToShape(attrs)
        self.view_port = None
        self.aliases = {}
        self.modules = []
        self.edges = []
        self.io_pins = {}
        self.nets = []
        self.groups = []

    def __str__(self):
        
        return '%s, %s, %s\n' '%s modules, %s io pins, %s nets, %s groups, %s edges\n' % (
            self.pos, self.shape, self.view_port,
            len(self.modules), len(self.io_pins), len(self.nets), len(self.groups), len(self.edges))

    def SetDefaultViewport(self):
        if not self.view_port:
            self.view_port = RectWHToRect4(self.pos, self.shape)

    def SetViewport(self, attrs):
        self.view_port = AttrsToBB(attrs)

    def AppendModule(self, attrs):
        self.modules.append((attrs[u'name'], Module(attrs)))

    def AppendEdge(self, attrs):
        self.edges.append(Edge(attrs))

    def AppendAlias(self, attrs):
        self.aliases[attrs[u'module']] = attrs[u'name']

    def AppendIOPin(self, attrs):
        self.io_pins[attrs[u'name']] = IOPin(attrs, self.pos, self.shape)

    def AppendNet(self, attrs):
        self.nets.append(Net(attrs))

    def AppendGroup(self, attrs):
        self.groups.append(Group(attrs))

    def FinishLoading(self):
        self.SetDefaultViewport()

        modules = self.modules
        self.modules = {}
        self.module_names = {}

        for name, module in modules:
            nx = module.pos[0]+self.pos[0]
            ny = module.pos[1]+self.pos[1]
            nllx = module.bb[0]+self.pos[0]
            nlly = module.bb[1]+self.pos[1]
            nurx = module.bb[2]+self.pos[0]
            nury = module.bb[3]+self.pos[1]
            module.pos = (nx, ny)
            module.bb = (nllx, nlly, nurx, nury)
            self.modules[name] = module
            try:
                self.module_names[self.aliases[name]] = name
            except KeyError:
                self.module_names[name] = name

class XMLDesign:

    class XMLDesignError:
        def __init__(self, msg):
            self.msg = msg

    def __init__(self, file):
        self.file = file
        self.stack = []

    def start_element(self, name, attrs):
        self.stack.append(name)

        if len(self.stack) == 1:
            if name != u'design':
                raise XMLDesign.XMLDesignError('expect design, get %s' % name)
            else:
                self.design = Design(attrs)
                return
        elif (len(self.stack) == 3) and (name == u'module'):
            if self.stack[1] != u'modules':
                raise XMLDesign.XMLDesignError('module not in modules')
            else:
                self.design.AppendModule(attrs)
        elif (len(self.stack) == 3) and (name == u'edge'):
            if self.stack[1] != u'edges':
                raise XMLDesign.XMLDesignError('edge not in edges')
            else:
                self.design.AppendEdge(attrs)
        elif (len(self.stack) == 3) and (name == u'alias'):
            if self.stack[1] != u'aliases':
                raise XMLDesign.XMLDesignError('alias not in aliases')
            else:
                self.design.AppendAlias(attrs)
        elif (len(self.stack) == 2) and (name == u'view_port'):
            self.design.SetViewport(attrs)

    def end_element(self, name):
        stored = self.stack.pop()
        if stored != name:
            raise XMLDesign.XMLDesignError('expect end %s, get %s' % (stored, name))

    def char_data(self, data):
        pass

    def load(self):
        assert not self.stack # load once

        p = xml.parsers.expat.ParserCreate()

        p.StartElementHandler = self.start_element
        p.EndElementHandler = self.end_element
        p.CharacterDataHandler = self.char_data

        try:
            try:
                f = open(self.file, 'r')

                wx.LogMessage('Processing %s...' % self.file)

                p.ParseFile(f)

                self.design.FinishLoading();

                if len(self.stack) != 0:
                    raise XMLDesign.XMLDesignError('bad XML')

                wx.LogMessage('Floorplan loaded.\n')

            finally:
                f.close()
        except XMLDesign.XMLDesignError, e:
            wx.LogError('%s: %s' % (e.msg, self.stack))
        except KeyError, e:
            wx.LogError('%s: %s' % (str(e), self.stack))
        except xml.parsers.expat.ExpatError, e:
            wx.LogError('%s: %s' % (str(e), self.stack))
        except IOError, e:
            wx.LogError(str(e))
        else:
            return True

class XMLLoadNets:

    class XMLLoadNetsError:
        def __init__(self, msg):
            self.msg = msg

    def __init__(self, file, design):
        self.file = file
        self.stack = []
        self.design = design

    def start_element(self, name, attrs):
        self.stack.append(name)

        if (len(self.stack) == 3) and (name == u'io_pin'):
            if self.stack[1] != u'io_pins':
                raise XMLDesign.XMLDesignError('io_pin not in io_pins')
            else:
                self.design.AppendIOPin(attrs)
        elif (len(self.stack) == 3) and (name == u'net'):
            if self.stack[1] != u'nets':
                raise XMLDesign.XMLDesignError('net not in nets')
            else:
                self.design.AppendNet(attrs)
        elif (len(self.stack) == 3) and (name == u'group'):
            if self.stack[1] != u'groups':
                raise XMLDesign.XMLDesignError('group not in groups')
            else:
                self.design.AppendGroup(attrs)

    def end_element(self, name):
        stored = self.stack.pop()
        if stored != name:
            raise XMLDesign.XMLDesignError('expect end %s, get %s' % (stored, name))

    def char_data(self, data):
        pass

    def load(self):
        assert not self.stack # load once

        p = xml.parsers.expat.ParserCreate()

        p.StartElementHandler = self.start_element
        p.EndElementHandler = self.end_element
        p.CharacterDataHandler = self.char_data

        try:
            try:
                f = open(self.file, 'r')

                wx.LogMessage('Processing %s...' % self.file)

                p.ParseFile(f)

                if len(self.stack) != 0:
                    raise XMLDesign.XMLDesignError('bad XML')

                wx.LogMessage('Nets loaded.\n')

            finally:
                f.close()
        except XMLLoadNets.XMLLoadNetsError, e:
            wx.LogError('%s: %s' % (e.msg, self.stack))
        except KeyError, e:
            wx.LogError('%s: %s' % (str(e), self.stack))
        except xml.parsers.expat.ExpatError, e:
            wx.LogError('%s: %s' % (str(e), self.stack))
        except IOError, e:
            wx.LogError(str(e))
        else:
            return True

class ChildView(wx.Window):

    def __init__(self, parent):
        wx.Window.__init__(self, parent,
            pos = (0, 0), size = (0, 0))

        self.Bind(wx.EVT_PAINT, self.OnPaint)
        self.Bind(wx.EVT_SIZE, self.OnSize)

        self.design_brush = wx.WHITE_BRUSH
        self.design_pen = wx.WHITE_PEN
        
        self.vp_brush = wx.Brush(wx.Colour(224, 224, 224))
        self.vp_pen = wx.Pen(wx.Colour(224, 224, 224))

        self.bb_pen = wx.LIGHT_GREY_PEN
        self.module_pen = wx.BLACK_PEN
        self.red_pen = wx.Pen(wx.Colour(128, 128, 128), style = wx.DOT)
        self.black_pen = wx.Pen(wx.Colour(128, 128, 128))
        self.net_pen = wx.Pen(wx.Colour(255, 0, 0))

        self.design = None
        
        self.ModuleRect4 = Module.PosRect4
        self.show_name = True
        self.show_alias = False
        self.show_edge = False
        self.show_bb = True
        self.show_net = False
        self.show_group = False
        self.show_group_location = False

    def OnPaint(self, evt):
        dc = wx.PaintDC(self)

        if not self.design:
            return

        self.CalcCanvas(self.GetClientSize(), (10,10))
        self.DrawDesign(dc)

    def OnSize(self, evt):
        self.Refresh()

    def OnViewModulePos(self, *event):
        if self.ModuleRect4 == Module.BBCenterRect4:
            self.ModuleRect4 = Module.PosRect4
        elif self.ModuleRect4 == Module.PosRect4:
            self.ModuleRect4 = Module.BBCenterRect4

        self.Refresh()

    def OnViewModuleName(self, *event):
        self.show_name = not self.show_name

        self.Refresh()

    def OnViewModuleAlias(self, *event):
        self.show_alias = not self.show_alias

        self.Refresh()

    def OnViewEdge(self, *event):
        self.show_edge = not self.show_edge

        self.Refresh()

    def OnViewBB(self, *event):
        self.show_bb = not self.show_bb

        self.Refresh()

    def OnViewNet(self, *event):
        self.show_net = not self.show_net

        self.Refresh()

    def OnViewGroup(self, *event):
        self.show_group = not self.show_group

        self.Refresh()

    def OnViewGroupLocation(self, *event):
        self.show_group_location = not self.show_group_location

        self.Refresh()

    def OnViewDesign(self, *event):
        if not self.design:
            return
        
        wx.LogMessage('Design Statistics:\n%s\n' % self.design)

    def UpdateDesign(self, design):
        self.design = design
        self.Refresh()

    def DrawDesign(self, dc):
        dc.SetBrush(self.vp_brush)
        dc.SetPen(self.vp_pen)
        self.DrawRect4(dc, *self.design.view_port)

        dc.SetBrush(self.design_brush)
#        dc.SetPen(self.design_pen)
        dc.SetPen(self.module_pen)
        self.DrawRect4(dc, *RectWHToRect4(self.design.pos, self.design.shape))

        dc.SetBrush(wx.TRANSPARENT_BRUSH)

        if self.show_group_location:
            if len(self.design.groups) <= 8:
                c, sc, cd = 2, 160, 40
            elif len(self.design.groups) <= 27:
                c, sc, cd = 3, 80, 40
            elif len(self.design.groups) <= 64:
                c, sc, cd = 4, 50, 50
            elif len(self.design.groups) <= 125:
                c, sc, cd = 5, 40, 40
            else:
                c, sc, cd = 6, 40, 20
                
            ci = c*c*c-1
                
            for group in self.design.groups:
                r = (ci%c)
                g = ((ci/c)%c)
                b = ((ci/c/c)%c)
                ci -= 1
                dc.SetBrush(wx.Brush(wx.Colour(r*sc+cd, g*sc+cd, b*sc+cd)))
                self.DrawRect4(dc, *group.bb)

            return

        if self.show_bb:
            dc.SetPen(self.bb_pen)
            for name, module in self.design.modules.items():
                self.DrawRect4(dc, *module.BBRect4())

        if self.show_group:
            if len(self.design.groups) <= 8:
                c, sc, cd = 2, 160, 40
            elif len(self.design.groups) <= 27:
                c, sc, cd = 3, 80, 40
            elif len(self.design.groups) <= 64:
                c, sc, cd = 4, 50, 50
            elif len(self.design.groups) <= 125:
                c, sc, cd = 5, 40, 40
            else:
                c, sc, cd = 6, 40, 20
                
            ci = c*c*c-1
                
            for group in self.design.groups:
                r = (ci%c)
                g = ((ci/c)%c)
                b = ((ci/c/c)%c)
                ci -= 1
                dc.SetBrush(wx.Brush(wx.Colour(r*sc+cd, g*sc+cd, b*sc+cd)))

#                wx.LogMessage('Group (%s,%s,%s) %s\n' % (r, g, b, group.modules))
                
                def GetModule(name):
                    localname = self.design.module_names[name]
                    return self.design.modules[localname]

                for module in map(GetModule, group.modules):
                    self.DrawRect4(dc, *self.ModuleRect4(module))

        dc.SetBrush(wx.TRANSPARENT_BRUSH)

        dc.SetPen(self.module_pen)
        for name, module in self.design.modules.items():
            self.DrawRect4(dc, *self.ModuleRect4(module))
            if self.show_name:
                if not self.show_alias:
                    self.DrawText4(dc, name, *self.ModuleRect4(module))
                else:
                    try:
                        self.DrawText4(dc, self.design.aliases[name], *self.ModuleRect4(module))
                    except KeyError:
                        self.DrawText4(dc, name, *self.ModuleRect4(module))

        if self.show_edge:
            for edge in self.design.edges:
                if edge.color == u'red':
                    dc.SetPen(self.red_pen)
                else:
                    dc.SetPen(self.black_pen)
                from_center = CenterRect4(*self.ModuleRect4(self.design.modules[edge.from_module]))
                to_center = CenterRect4(*self.ModuleRect4(self.design.modules[edge.to_module]))
                self.DrawArrow(dc, from_center, to_center)

        if self.show_net:
            for net in self.design.nets:
                dc.SetPen(self.net_pen)
                def GetLocation(name):
                    try:
                        n = self.design.module_names[name]
                        return CenterRect4(*self.ModuleRect4(self.design.modules[n]))
                    except KeyError:
                        return self.design.io_pins[name].pos

                pin_locations = [GetLocation(name) for name in net.pins]
                if len(pin_locations) == 2:
                    self.DrawArrow(dc, pin_locations[0], pin_locations[1])
                else:
                    cx, cy = 0, 0
                    for x, y in pin_locations:
                        cx += x
                        cy += y
                    cx /= len(pin_locations)
                    cy /= len(pin_locations)
                    for pos in pin_locations:
                        self.DrawArrow(dc, pos, (cx, cy))

    def CalcCanvas(self, client, blank):
        bx, by = blank
        cw, ch = client

        llx, lly, urx, ury = self.design.view_port

        sx = float(cw-2*bx)/(urx-llx)
        sy = float(ch-2*bx)/(ury-lly)

        if sx > sy:
            self.sx = sy
            self.sy = -sy
        else:
            self.sx = sx
            self.sy = -sx
        
        self.dx = float(cw)/2-(llx+urx)*self.sx/2
        self.dy = float(ch)/2-(lly+ury)*self.sy/2

    def CalcCanvasPS(self, client):
        cw, ch = client

        llx, lly, urx, ury = self.design.view_port

        sx = float(cw)/(urx-llx)
        sy = float(ch)/(ury-lly)

        if sx > sy:
            self.sx = sy
            self.sy = -sy
        else:
            self.sx = sx
            self.sy = -sx

        self.dx = -llx*self.sx
        self.dy = ch-lly*self.sy

    def PointToClient(self, *pt):
        return (long(pt[0]*self.sx+self.dx), long(pt[1]*self.sy+self.dy))

    def PointFromClient(self, *pt):
        return ((pt[0]-self.dx)/self.sx, (pt[1]-self.dy)/self.sy)

    def Rect4ToClient(self, *rc):
        x0, y0, x1, y1 = rc
        cx0, cy0 = self.PointToClient(x0, y0)
        cx1, cy1 = self.PointToClient(x1, y1)
        return (cx0, cy0, cx1, cy1)

    def DrawRect4(self, dc, *rc):
        pos, shape = Rect4ToRectWH(*self.Rect4ToClient(*rc))
        dc.DrawRectangle(pos[0], pos[1], shape[0], shape[1])

    def DrawText4(self, dc, str, *rc):
        w, h = dc.GetTextExtent(str)
        llx, lly, urx, ury = self.Rect4ToClient(*rc)
        if abs(urx-llx) > w*0.8 and abs(ury-lly) > h*0.8:
            dc.DrawText(str, (llx+urx-w)/2, (lly+ury-h)/2)
        elif abs(urx-llx) > h*0.8 and abs(ury-lly) > w*0.8:
            dc.DrawRotatedText(str, (llx+urx+h)/2, (lly+ury-w)/2, 270)
        else:
            pass

    def DrawArrow(self, dc, pt0, pt1):
        x0, y0 = self.PointToClient(*pt0)
        x1, y1 = self.PointToClient(*pt1)
        dc.DrawLine(x0, y0, x1, y1)

class MainFrame(wx.Frame):

    def __init__(self, parent, title):
        wx.Frame.__init__(self, parent, -1, title)

        self.splitter = wx.SplitterWindow(self, -1,
            style = wx.CLIP_CHILDREN|wx.SP_LIVE_UPDATE|wx.SP_3D)

        self.child = ChildView(self.splitter)

        self.log_view = wx.TextCtrl(self.splitter, -1,
            style = wx.TE_MULTILINE|wx.TE_READONLY|wx.VSCROLL|wx.NO_BORDER)    

        self.splitter.SplitVertically(self.child, self.log_view, 600)
        self.splitter.SetMinimumPaneSize(50)

        class Log(wx.PyLog):

            def __init__(self, textCtrl):
                wx.PyLog.__init__(self)
                self.tc = textCtrl

            def DoLogString(self, message, timestamp):
                if self.tc:
                    self.tc.AppendText(message + '\n')

        self.log = Log(self.log_view) 
        wx.Log_SetActiveTarget(self.log)

        self.CreateStatusBar()

        self.CreateMenuBar()

        if '__WXMSW__' in wx.PlatformInfo:
            self.Maximize()

        self.design = None

    def CreateMenuBar(self):
        self.menu = wx.MenuBar()

        for name, list in [
            ('&File', [
                (self.OnFileOpen, '&Open...\tCtrl-O',
                    'Open a new floorplan.'),
                (self.OnFileLoadNets, '&Load AUX file...\tCtrl-L',
                    'Load interconnections.'),
                (self.OnRefresh, '&Refresh\tF5',
                    'Refresh the current floorplan.'),
                (self.OnFilePS, '&Export PS...',
                    'Export the current floorplan in PostScript format.'),
                (self.OnPrint, '&Print...\tCtrl-P',
                    'Print the current floorplan.'),
                (lambda *event: self.Close(), 'E&xit\tAlt-X',
                    'Get the heck outta here!'),
                ]),
            ('&View', [
                (self.child.OnViewDesign, 'Show Design Statistics',
                    'Click to show design information.'),
                (self.child.OnViewModulePos, 'Place Modules by BB\tCtrl-C',
                    'Place modules at the center of their bounding boxes if checked.', wx.ITEM_CHECK),
                (self.child.OnViewModuleName, 'Hide Module Names\tCtrl-N',
                    'Hide module names if checked.', wx.ITEM_CHECK),
                (self.child.OnViewModuleAlias, 'Show Module Alias\tCtrl-A',
                    'Show module alias instead of name if checked.', wx.ITEM_CHECK),
                (self.child.OnViewEdge, 'Show Edges\tCtrl-E',
                    'Show graph edges if checked.', wx.ITEM_CHECK),
                (self.child.OnViewBB, 'Hide BB\tCtrl-B',
                    'Hide bounding boxes if checked.', wx.ITEM_CHECK),
                (self.child.OnViewNet, 'Show Nets\tCtrl-S',
                    'Show interconnect nets if checked.', wx.ITEM_CHECK),
                (self.child.OnViewGroup, 'Show Groups\tCtrl-G',
                    'Show groups in different colors.', wx.ITEM_CHECK),
                (self.child.OnViewGroupLocation, 'Show Group Locations\tCtrl-R',
                    'Show groups in their original locations with colors.', wx.ITEM_CHECK),
                ]),
            ('&Help', [
                (self.OnHelpAbout, '&About',
                    'About %s' % g_title),
                ]),
            ]:

            menu = wx.Menu()

            for item in list:
                self.Bind(wx.EVT_MENU, item[0], menu.Append(-1, *item[1:]))

            self.menu.Append(menu, name)

#        wx.App_SetMacExitMenuItemId(item.GetId())

        self.SetMenuBar(self.menu)

    def OnHelpAbout(self, *event):
        dlg = wx.MessageDialog(self,
                               message = '%s v%s by %s.' % (g_title, g_version, g_author),
                               caption = 'About',
                               style = wx.OK|wx.ICON_INFORMATION)
        dlg.ShowModal()
        dlg.Destroy()

    def OnFileOpen(self, *event):
        dlg = wx.FileDialog(self, message = "Open a new floorplan",
            wildcard = "XML files (*.xml)|*.xml|All files (*.*)|*.*",
            style = wx.OPEN|wx.CHANGE_DIR)

        if dlg.ShowModal() == wx.ID_OK:
            file = dlg.GetPath()

            design = self.ReadXMLDesign(file)

            if design:
                self.SetTitle('%s - %s (%s, %s)' % (g_title, file, design.shape[0], design.shape[1]))
                self.design = (file, design)
                self.child.UpdateDesign(self.design[1])

        dlg.Destroy()

    def OnFileLoadNets(self, *event):
        if not self.design:
            return
        
        dlg = wx.FileDialog(self, message = "Load interconnects and groups",
            wildcard = "XML files (*.xml)|*.xml|All files (*.*)|*.*",
            style = wx.OPEN|wx.CHANGE_DIR)

        if dlg.ShowModal() == wx.ID_OK:
            file = dlg.GetPath()

            load_nets = XMLLoadNets(file, self.design[1])
            load_nets.load()
            self.child.Refresh()

        dlg.Destroy()

    def OnFilePS(self, *event):
        if not self.design:
            return

        dlg = wx.FileDialog(self, message = "Export PS",
            wildcard = "PostScript files (*.ps)|*.ps",
            style = wx.SAVE|wx.OVERWRITE_PROMPT|wx.CHANGE_DIR)

        if dlg.ShowModal() == wx.ID_OK:
            file = dlg.GetPath()

            data = wx.PrintData()
            data.SetFilename(file)
            data.SetPaperId(wx.PAPER_LETTER)

            dc = wx.PostScriptDC(data)

            if dc.Ok():
                dc.StartDoc('Floorplan')
                dc.StartPage()
                dc.SetFont(wx.NORMAL_FONT)
                self.child.CalcCanvasPS(dc.GetSizeTuple())
                self.child.DrawDesign(dc)
                dc.EndPage()
                dc.EndDoc()

                wx.LogMessage('%s exported.' % file)

            else:
                err = wx.MessageDialog(self,
                                       message = 'Fail to export as %s!' % file,
                                       caption = 'Error',
                                       style = wx.OK|wx.ICON_ERROR)
                err.ShowModal()
                err.Destroy()

        dlg.Destroy()

    def OnPrint(self, *event):
        if not self.design:
            return

        dlg = wx.PrintDialog(self)

        if dlg.ShowModal() == wx.ID_OK:

            dc = dlg.GetPrintDC()

            if dc.Ok():
                dc.StartDoc('Floorplan')
                dc.StartPage()
                dc.SetFont(wx.NORMAL_FONT)
                self.child.CalcCanvasPS(dc.GetSizeTuple())
                self.child.DrawDesign(dc)
                dc.EndPage()
                dc.EndDoc()

                wx.LogMessage('Printed.')

            else:
                err = wx.MessageDialog(self,
                                       message = 'Fail to print!',
                                       caption = 'Error',
                                       style = wx.OK|wx.ICON_ERROR)
                err.ShowModal()
                err.Destroy()

        dlg.Destroy()

    def OnRefresh(self, *event):
        if not self.design:
            return
        
        file = self.design[0]

        design = self.ReadXMLDesign(file)

        if design:
            self.SetTitle('%s - %s (%s, %s)' % (g_title, file, design.shape[0], design.shape[1]))
            self.design = (file, design)
            self.child.UpdateDesign(self.design[1])

    def ReadXMLDesign(self, file):
        xml_design = XMLDesign(file)
        
        if xml_design.load():
            return xml_design.design
        else:
            dlg = wx.MessageDialog(self,
                                   message = 'Fail to read %s!' % file,
                                   caption = 'Error',
                                   style = wx.OK|wx.ICON_ERROR)
            dlg.ShowModal()
            dlg.Destroy()

def main():
    
    class MyApp(wx.App):

        def OnInit(self):
            frame = MainFrame(None, g_title)
            frame.Show()
            return True

    app = MyApp(False)
    app.MainLoop()

if __name__ == '__main__':
    main()
